#include <ctype.h>
#include <errno.h>
#include <stdbool.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <sys/types.h>
#include <afblib/inbuf.h>
#include <afblib/inbuf_scan.h>
#include <afblib/outbuf.h>
#include <afblib/outbuf_printf.h>
#include <stralloc.h>
#include "tf-session.h"

/* emit HTTP error response onto out */
void fail(outbuf* out, const char* message) {
   outbuf_printf(out,
      "HTTP/1.0 500 Failed\r\n"
      "\r\n"
      "%s\r\n",
      message
   );
   outbuf_flush(out);
}

/* both images are in the public domain, no attribution necessary */
static char frog_url[] =
   "https://upload.wikimedia.org/wikipedia/commons/f/f1/Frog_from_Roesel_von_Rosenhof%3B_1758_Wellcome_L0001704.jpg";
static char toad_url[] = "https://upload.wikimedia.org/wikipedia/commons/9/98/Frog_from_Roesel_von_Rosenhof%3B_1758_Wellcome_L0001710.jpg";

typedef enum {MOVES_POSSIBLE, GAME_STUCK, GAME_WON} GameStatus;

GameStatus print_status(outbuf* out, stralloc* tf_status) {
   unsigned int len = tf_status->len;
   outbuf_printf(out, "<table border=\"1\" style=\"table-layout:fixed\"><tr>\r\n");
   unsigned int width = 100 / len;
   enum {IN_TOADS, IN_SPACE, IN_FROGS, UNFINISHED} status = IN_TOADS;
   stralloc next = {0}; unsigned int possible_moves = 0;
   for (unsigned int i = 0; i < len; ++i) {
      /* state machine that checks if the game is finished */
      switch (tf_status->s[i]) {
	 case 'T': if (status != IN_TOADS) status = UNFINISHED; break;
	 case 'E': if (status == IN_TOADS) status = IN_SPACE; break;
	 case 'F':
	    if (status == IN_SPACE)
	       status = IN_FROGS;
	    else if (status != IN_FROGS) {
	       status = UNFINISHED;
	    }
	    break;
      };
      /* check, if a move is possible, and, if yes, compute the next state */
      next.len = 0;
      if (tf_status->s[i] == 'F' && i + 1 < len) {
	 if (tf_status->s[i+1] == 'E') {
	    stralloc_copy(&next, tf_status);
	    next.s[i] = 'E';
	    next.s[i+1] = 'F';
	    ++possible_moves;
	 } else if (tf_status->s[i+1] == 'T' && i + 2 < len &&
	       tf_status->s[i+2] == 'E') {
	    stralloc_copy(&next, tf_status);
	    next.s[i] = 'E';
	    next.s[i+2] = 'F';
	    ++possible_moves;
	 }
      } else if (tf_status->s[i] == 'T' && i > 0) {
	 if (tf_status->s[i-1] == 'E') {
	    stralloc_copy(&next, tf_status);
	    next.s[i] = 'E';
	    next.s[i-1] = 'T';
	    ++possible_moves;
	 } else if (tf_status->s[i-1] == 'F' && i > 1 &&
	       tf_status->s[i-2] == 'E') {
	    stralloc_copy(&next, tf_status);
	    next.s[i] = 'E';
	    next.s[i-2] = 'T';
	    ++possible_moves;
	 }
      }
      outbuf_printf(out, "   <td align=center style=\"width:%d%%\">", width);
      if (tf_status->s[i] == 'E') {
	 outbuf_printf(out, "&nbsp;");
      } else {
	 if (next.len > 0) {
	    stralloc_0(&next);
	    outbuf_printf(out, "<a href=\"/%s\">", next);
	 }
	 if (tf_status->s[i] == 'F') {
	    outbuf_printf(out, "<img alt=\"frog\" width=\"100\%\" src=\"%s\">",
	       frog_url);
	 } else {
	    outbuf_printf(out, "<img alt=\"toad\" width=\"100\%\" src=\"%s\">",
	       toad_url);
	 }
	 if (next.len > 0) {
	    outbuf_printf(out, "</a>");
	 }
      }
      outbuf_printf(out, "</td>\r\n");
   }
   outbuf_printf(out, "</tr></table>\r\n");
   stralloc_free(&next);
   if (possible_moves) {
      return MOVES_POSSIBLE;
   } else if (status == UNFINISHED) {
      return GAME_STUCK;
   } else {
      return GAME_WON;
   }
}

void tf_session(int fd) {
   inbuf in = {fd};
   outbuf out = {fd};

   /* parse incoming HTTP request */
   stralloc tf_status = {0};
   int count = inbuf_scan(&in,
      /* HTTP GET command */
      "GET"
      " "
      /* optional URL with host and port */
      "(?:"
         "http://[-.a-zA-Z90-9]+"
	 "(?:\\d+)?"
      ")?"
      "/"
      "(?:"
	 "([EFT]+)"  /* game status */
      ")?"
      " "
      "HTTP/1\\.[01]"
      "\r\n",
      &tf_status
   );
   if (count != 0 && count != 1) {
      fail(&out, "Unable to parse your HTTP request");
      return;
   }
   if (count == 0) {
      /* determine length of our field */
      {
	 static bool seeded = false;
	 if (!seeded) {
	    srand(getpid() ^ time(0));
	    seeded = true;
	 }
      }
      unsigned int frogs = 2 + rand() % 3;
      unsigned int toads = 2 + rand() % 3;
      for (unsigned int i = 0; i < frogs; ++i) {
	 stralloc_cats(&tf_status, "F");
      }
      stralloc_cats(&tf_status, "E");
      for (unsigned int i = 0; i < toads; ++i) {
	 stralloc_cats(&tf_status, "T");
      }
   }
   unsigned int len = tf_status.len;

   /* deliver regular HTTP response */
   outbuf_printf(&out,
      "HTTP/1.0 200 Ok\r\n"
      "Content-Type:text/html\r\n"
      "\r\n"
      "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\"\r\n"
      "   \"http://www.w3.org/TR/html4/loose.dtd\">\r\n"
      "<html><head>\r\n"
      "<title>Toads and Frogs</title>"
      "<meta http-equiv=\"Content-Type\""
      " content=\"text/html; charset=iso-8859-1\">"
      "</head><body>\r\n"
      "<h1>*** Toads and Frogs ***</h1>\r\n"
   );
   GameStatus status = print_status(&out, &tf_status);
   switch (status) {
      case MOVES_POSSIBLE: /* no further output */ break;
      case GAME_STUCK:
	 outbuf_printf(&out, "<p>You are stuck, no more moves possible!</p>"
	    "<p><a href=\"/\">New game?</a></p>");
	 break;
      case GAME_WON:
	 outbuf_printf(&out, "<p>Congratulations, you've won!</p>"
	    "<p><a href=\"/\">New game?</a></p>");
	 break;
   }
   /*
   outbuf_printf(&out, "<p>"
      "<a href=\"http://validator.w3.org/check?uri=referer\">"
      "<img src=\"http://www.w3.org/Icons/valid-html401\" "
	 "alt=\"Valid HTML 4.01 Transitional\" height=\"31\" width=\"88\"></a>"
      "</p>");
   */
   outbuf_printf(&out, "</body></html>\r\n");
   outbuf_flush(&out);
   stralloc_free(&tf_status);
}
